/** Decode a "floating point" parameter from the special bit pattern
    format. Used in off-line storage (generated by the tool programs);
    not compatible with MIDI SysEx.

  This code (and algorithm) is now localized here; it must match the
  encoding code in synti2_fenc.c

  Format:
  - Input is an integer (actually a MIDI "varlength" of 1-4 bytes, 
    but we assume it is read into an int before calling this function)
  - two least significant bits are the accuracy, as negative power of ten.
  - shift right to rid of two bits.
  - the next bit is the sign (1 = positive, 0 = negative)
    FIXME: Alternatives for this?
  - shift right to rid of one bit.
  - the rest of bits are now the significant digits of the value
  - apply sign and the negative power of ten (0-3)

  - examples of one-byte parameters: 0.001 0.015 0.15 -12
  - examples of two-byte parameters: 0.016 -0.21 19 100

 */

float synti2_decode_f(int stored){
  int i;
  float res;
  int positive;
  i = stored & 0x3; stored >>= 2;
  positive = stored & 0x1; stored >>= 1;
  res = ((positive)?stored:-stored);
  for (; i > 0; i--) res /= 10.f;
  return res;
}
